//----------------------------------------------------------------------------
// Copyright 2008-2023 Joe Lowe
//
// Permission is granted to any person obtaining a copy of this Software,
// to deal in the Software without restriction, including the rights to use,
// copy, modify, merge, publish, distribute, sublicense, and sell copies of
// the Software.
//
// The above copyright and permission notice must be left intact in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS WITHOUT WARRANTY.
//----------------------------------------------------------------------------
// file name:  bitops.h
// created:    2008.10.07
//
// Compiler specific wrappers/implementations for performant integer bit
// manipulation.
//
// Rotate value right/left by count bits.
//   unsigned bits_ror(unsigned v,unsigned c)
//   uintptr_t bits_rorp(uintptr_t v,unsigned c)
//   uint64_t bits_ror64(uint64_t v,unsigned c)
//   unsigned bits_rol(unsigned v,unsigned c)
//   uintptr_t bits_rolp(uintptr_t v,unsigned c)
//   uint64_t bits_rol64(uint64_t v,unsigned c)
// Arithmetic shift signed value right, preserving sign bit.
//   int bits_sar(int v,unsigned c)
//   intptr_t bits_sarp(intptr_t v,unsigned c)
//   int64_t bits_sar64(int64_t v,unsigned c)
// Shift signed value left, overflow destroying sign bit.
//   int bits_sal(int v,unsigned c)
//   intptr_t bits_salp(intptr_t v,unsigned c)
//   int64_t bits_sal64(int64_t v,unsigned c)
// Return count of sequential zero bits starting at high bit.
//   unsigned bits_clz(unsigned v)
//   unsigned bits_clzp(uintptr_t v)
//   unsigned bits_clz64(uint64_t v)
// Return count of sequential zero bits starting at low bit.
//   unsigned bits_ctz(unsigned v)
//   unsigned bits_ctzp(uintptr_t v)
//   unsigned bits_ctz64(uint64_t v)
// Return one-based index of highest non zero bit. Returns zero if all
// bits are zero.
//   unsigned bits_fls(unsigned v)
//   unsigned bits_flsp(uintptr_t v)
//   unsigned bits_fls64(uint64_t v)
// Return one-based index of lowest non zero bit. Returns zero if all
// bits are zero.
//   unsigned bits_ffs(unsigned v)
//   unsigned bits_ffsp(uintptr_t v)
//   unsigned bits_ffs64(uint64_t v)
// Return count of non-zero bits.
//   unsigned bits_pop(unsigned v)
//   unsigned bits_popp(uintptr_t v)
//   unsigned bits_pop64(uint64_t v)
// Byte swap.
//   uint16_t bits_swap16(uint16_t v)
//   unsigned bits_swap32(unsigned v)
//   uintptr_t bits_swapp(uintptr_t v)
//   uint64_t bits_swap64(uint64_t v)
// Double precision unsigned multiply.
//   unsigned bits_umul(unsigned a,unsigned b,unsigned* h)
//   uintptr_t bits_umulp(uintptr_t a,uintptr_t b,uintptr_t* h)
//   uint64_t bits_umul64(uint64_t a,uint64_t b,uint64_t* h)
//   unsigned bits_usqr(unsigned a,unsigned* h)
//   uintptr_t bits_usqrp(uintptr_t a,uintptr_t* h)
//   uint64_t bits_usqr64(uint64_t a,uint64_t* h)
// Double precision unsigned divide.
//   unsigned bits_udiv(unsigned nl,unsigned nh,unsigned d)
//   uintptr_t bits_udivp(uintptr_t nl,uintptr_t nh,uintptr_t d)
//   uint64_t bits_udiv64(uint64_t nl,uint64_t nh,uint64_t d)
//   unsigned bits_umod(unsigned nl,unsigned nh,unsigned d)
//   uintptr_t bits_umodp(uintptr_t nl,uintptr_t nh,uintptr_t d)
//   uint64_t bits_umod64(uint64_t nl,uint64_t nh,uint64_t d)
//   unsigned bits_udivmod(unsigned nl,unsigned nh,unsigned d,unsigned* r)
//   uintptr_t bits_udivmodp(uintptr_t nl,uintptr_t nh,uintptr_t d,uintptr_t* r)
//   uint64_t bits_udivmod64(uint64_t nl,uint64_t nh,uint64_t d,uint64_t* r)
//----------------------------------------------------------------------------
#ifndef BITOPS_H
#define BITOPS_H
#include "ptcompiler.h"
PT_EXTERNC_START

   // Technically this should not be a constant, or
   // should vary based on architecture.
PT_STATIC_CONST(uintptr_t,bits_cache_line_size,64);

   // These rotate functions are expected to be recognized by
   // the clang/gcc and optimized down to single instructions
   // in many cases.
PT_INLINE unsigned bits_ror_slow( unsigned v, unsigned c) { c &= 31; return (v >> c) | (v << (32 - c)); }
PT_INLINE unsigned bits_rol_slow( unsigned v, unsigned c) { c &= 31; return (v << c) | (v >> (32 - c)); }
PT_INLINE uint64_t bits_ror64_slow( uint64_t v, unsigned c) { c &= 63; return (v >> c) | (v << (64 - c)); }
PT_INLINE uint64_t bits_rol64_slow( uint64_t v, unsigned c) { c &= 63; return (v << c) | (v >> (64 - c)); }

static unsigned bits_clz_slow_(unsigned v)
{
   unsigned c; PT_ASSERT(v);
   c = PT_SCAST( unsigned, (v & 0xFFFF0000) == 0) << 4; c += PT_SCAST( unsigned, (v & (0xFF000000u >> c)) == 0) << 3;
   c += PT_SCAST( unsigned, (v & (0xF0000000u >> c)) == 0) << 2; c += PT_SCAST( unsigned, (v & (0xC0000000u >> c)) == 0) << 1;
   return c + ((v & (0x80000000u >> c)) == 0);
}

static unsigned bits_ctz_slow_(unsigned v)
{
   unsigned c; PT_ASSERT(v);
   c = PT_SCAST( unsigned, (v & 0x0000FFFF) == 0) << 4; c += PT_SCAST( unsigned, (v & (0x000000FFu << c)) == 0) << 3;
   c += PT_SCAST( unsigned, (v & (0x0000000Fu << c)) == 0) << 2; c += PT_SCAST( unsigned, (v & (0x00000003u << c)) == 0) << 1;
   return c + ((v & (0x00000001u << c)) == 0);
}

static unsigned bits_pop_slow_(unsigned v)
{
   v = (v&0x55555555)+((v>>1)&0x55555555);
   v = (v&0x33333333)+((v>>2)&0x33333333);
   v = (v&0x0F0F0F0F)+((v>>4)&0x0F0F0F0F);
   v += v>>8; return PT_SCAST(uint8_t,v+(v>>16));
}

#if SIZE_MAX > UINT_MAX
static unsigned bits_pop64_slow_( uint64_t v)
{
   v = (v&0x5555555555555555)+((v>>1)&0x5555555555555555);
   v = (v&0x3333333333333333)+((v>>2)&0x3333333333333333);
   v = (v&0x0F0F0F0F0F0F0F0F)+((v>>4)&0x0F0F0F0F0F0F0F0F);
   v += v>>8; v += v>>16; return PT_SCAST(uint8_t,v+(v>>32));
}
#endif

PT_INLINE unsigned bits_clz64_slow_(uint64_t v) { unsigned h = PT_SCAST(unsigned,v>>32); PT_ASSERT(v); return h?bits_clz_slow_(h):(32U+bits_clz_slow_(PT_SCAST(unsigned,v))); }
PT_INLINE unsigned bits_ctz64_slow_(uint64_t v) { PT_ASSERT(v); return PT_SCAST(unsigned,v)?bits_ctz_slow_(PT_SCAST(unsigned,v)):(32U+bits_ctz_slow_(PT_SCAST(unsigned,v>>32U))); }

PT_INLINE unsigned bits_clz_slow(unsigned v) { return v?bits_clz_slow_(v):32U; }
PT_INLINE unsigned bits_ctz_slow(unsigned v) { return v?bits_ctz_slow_(v):32U; }
PT_INLINE unsigned bits_fls_slow(unsigned v) { return v?(32U-bits_clz_slow_(v)):0U; }
PT_INLINE unsigned bits_ffs_slow(unsigned v) { return v?(bits_ctz_slow_(v)+1U):0U; }
PT_INLINE unsigned bits_pop_slow(unsigned v) { return bits_pop_slow_(v); }
PT_INLINE unsigned bits_clz64_slow(uint64_t v) { return v?bits_clz64_slow_(v):64U; }
PT_INLINE unsigned bits_ctz64_slow(uint64_t v) { return v?bits_ctz64_slow_(v):64U; }
PT_INLINE unsigned bits_fls64_slow(uint64_t v) { return v?(64U-bits_clz64_slow_(v)):0U; }
PT_INLINE unsigned bits_ffs64_slow(uint64_t v) { return v?(bits_ctz64_slow_(v)+1U):0U; }
PT_INLINE unsigned bits_pop64_slow(uint64_t v) { return bits_pop_slow_(PT_SCAST(unsigned,v))+bits_pop_slow_(PT_SCAST(unsigned,v>>32)); }
PT_INLINE uint16_t bits_swap16_slow( uint16_t v) { return PT_SCAST( uint16_t, (PT_SCAST( unsigned, v) >> 8) | (PT_SCAST( unsigned, v) << 8)); }
PT_INLINE unsigned bits_swap32_slow( unsigned v) { return ((v>>24)&0x00FF)|((v>>8)&0xFF00)|((v&0xFF00)<<8)|(v<<24); }
PT_INLINE uint64_t bits_swap64_slow(uint64_t v) { return ((v>>56)&0x000000FF)|((v>>40)&0x0000FF00)|((v>>24)&0x00FF0000)|((v>>8)&0xFF000000)|((v&0xFF000000)<<8)|((v&0x00FF0000)<<24)|((v&0x0000FF00)<<40)|(v<<56); }

PT_INLINE unsigned bits_umul_slow(unsigned a,unsigned b,void* h)
{
   uint64_t p = PT_SCAST(uint64_t,a)*PT_SCAST(uint64_t,b);
   *PT_SCASTP(unsigned*,h) = PT_SCAST(unsigned,p>>32);
   return PT_SCAST(unsigned,p);
}

PT_INLINE unsigned bits_usqr_slow(unsigned a,void* h)
{
   uint64_t p = PT_SCAST(uint64_t,a)*PT_SCAST(uint64_t,a);
   *PT_SCASTP(unsigned*,h) = PT_SCAST(unsigned,p>>32);
   return PT_SCAST(unsigned,p);
}

PT_INLINE uint64_t bits_umul64_slow(uint64_t a,uint64_t b,void* h)
{
   uint64_t albh = (a&0xFFFFFFFFU)*(b>>32);
   uint64_t ahbl = (a>>32)*(b&0xFFFFFFFFU);
   uint64_t albl = (a&0xFFFFFFFFU)*(b&0xFFFFFFFFU);
   uint64_t d = albl+(albh<<32);
   uint64_t c = d+(ahbl<<32);
   *PT_SCASTP(uint64_t*,h) = (a>>32)*(b>>32)+(albh>>32)+(ahbl>>32)+(d<albl)+(c<d);
   return c;
}

PT_INLINE uint64_t bits_usqr64_slow(uint64_t a,void* h)
{
   uint64_t lh = (a&0xFFFFFFFFU)*(a>>32);
   uint64_t ll = (a&0xFFFFFFFFU)*(a&0xFFFFFFFFU);
   uint64_t d = ll+(lh<<32);
   uint64_t c = d+(lh<<32);
   *PT_SCASTP(uint64_t*,h) = (a>>32)*(a>>32)+(lh>>32)+(lh>>32)+(d<ll)+(c<d);
   return c;
}

PT_INLINE unsigned bits_udiv_slow(unsigned nl,unsigned nh,unsigned d)
{
   uint64_t n = (PT_SCAST(uint64_t,nh)<<32)|nl;
   return PT_SCAST(unsigned,n/d);
}

PT_INLINE unsigned bits_umod_slow(unsigned nl,unsigned nh,unsigned d)
{
   uint64_t n = (PT_SCAST(uint64_t,nh)<<32)|nl;
   return PT_SCAST(unsigned,n%d);
}

PT_INLINE unsigned bits_udivmod_slow(unsigned nl,unsigned nh,unsigned d,void* r)
{
   uint64_t n = (PT_SCAST(uint64_t,nh)<<32)|nl;
   *PT_SCASTP(unsigned*,r) = PT_SCAST(unsigned,n%d);
   return PT_SCAST(unsigned,n/d);
}

uint64_t PT_CCALL bits_udivmod64_slow(uint64_t nl,uint64_t nh,uint64_t d,void* r);

PT_INLINE uint64_t bits_udiv64_slow(uint64_t nl,uint64_t nh,uint64_t d)
{
   return bits_udivmod64_slow(nl,nh,d,nullptr);
}

PT_INLINE uint64_t bits_umod64_slow(uint64_t nl,uint64_t nh,uint64_t d)
{
   uint64_t r;
   bits_udivmod64_slow(nl,nh,d,&r);
   return r;
}

#ifdef _MSC_VER

 #define BITS_ROT_DEFINED_
 unsigned int __cdecl _rotr(unsigned int,int);
 #pragma intrinsic(_rotr)
 unsigned int __cdecl _rotl(unsigned int,int);
 #pragma intrinsic(_rotl)
 PT_INLINE unsigned bits_ror(unsigned v,unsigned c) { return _rotr(v,PT_SCAST(int,c)); }
 PT_INLINE unsigned bits_rol(unsigned v,unsigned c) { return _rotl(v,PT_SCAST(int,c)); }
 #if UINTPTR_MAX > UINT_MAX
  unsigned __int64 _rotr64(unsigned __int64,int);
  #pragma intrinsic(_rotr64)
  unsigned __int64 _rotl64(unsigned __int64,int);
  #pragma intrinsic(_rotl64)
  #define BITS_ROT64_DEFINED_
  PT_INLINE uintptr_t bits_ror64(uint64_t v,unsigned c) { return _rotr64(v,PT_SCAST(int,c)); }
  PT_INLINE uintptr_t bits_rol64(uint64_t v,unsigned c) { return _rotl64(v,PT_SCAST(int,c)); }
 #endif

  // Bit scan intrinsic in MS compiler used to implement clz/fls/ctz/ffs.
 #define BITS_CLZ_DEFINED_
 #ifndef BitScanReverse
  unsigned char _BitScanReverse(unsigned long*,unsigned long);
  #pragma intrinsic(_BitScanReverse)
  unsigned char _BitScanForward(unsigned long*,unsigned long);
  #pragma intrinsic(_BitScanForward)
 #endif
 PT_INLINE unsigned bits_clz(unsigned v) { unsigned long i; unsigned f = _BitScanReverse(&i,v); return PT_SCAST(unsigned,sizeof(v)*8)-((0U-f)&(PT_SCAST(unsigned,i)+1U)); }
 PT_INLINE unsigned bits_fls(unsigned v) { unsigned long i; unsigned f = _BitScanReverse(&i,v); return (0U-f)&(PT_SCAST(unsigned,i)+1U); }
 PT_INLINE unsigned bits_ctz(unsigned v) { unsigned long i; unsigned f = _BitScanForward(&i,v); return ((0U-f)&PT_SCAST(unsigned,i))|((UINT_MAX+f)&PT_SCAST(unsigned,sizeof(v)*8)); }
 PT_INLINE unsigned bits_ffs(unsigned v) { unsigned long i; unsigned f = _BitScanForward(&i,v); return (0U-f)&(PT_SCAST(unsigned,i)+1U); }
 #if UINTPTR_MAX > UINT_MAX
  #ifndef BitScanReverse64
   unsigned char _BitScanReverse64(unsigned long*,uint64_t);
   #pragma intrinsic(_BitScanReverse64)
   unsigned char _BitScanForward64(unsigned long*,uint64_t);
   #pragma intrinsic(_BitScanForward64)
  #endif
  PT_INLINE unsigned bits_clz64(uint64_t v) { unsigned long i; unsigned f = _BitScanReverse64(&i,v); return PT_SCAST(unsigned,sizeof(v)*8)-((0U-f)&(PT_SCAST(unsigned,i)+1U)); }
  PT_INLINE unsigned bits_fls64(uint64_t v) { unsigned long i; unsigned f = _BitScanReverse64(&i,v); return (0U-f)&(PT_SCAST(unsigned,i)+1U); }
  PT_INLINE unsigned bits_ctz64(uint64_t v) { unsigned long i; unsigned f = _BitScanForward64(&i,v); return ((0U-f)&PT_SCAST(unsigned,i))|((UINT_MAX+f)&PT_SCAST(unsigned,sizeof(v)*8)); }
  PT_INLINE unsigned bits_ffs64(uint64_t v) { unsigned long i; unsigned f = _BitScanForward64(&i,v); return (0U-f)&(PT_SCAST(unsigned,i)+1U); }
 #else
  PT_INLINE unsigned bits_clz64(uint64_t v) { unsigned h = PT_SCAST(unsigned,v>>32); return h?bits_clz(h):(32U+bits_clz(PT_SCAST(unsigned,v))); }
  PT_INLINE unsigned bits_fls64(uint64_t v) { return v?PT_SCAST(unsigned,sizeof(v)*8U)-bits_clz64(v):0U; }
  PT_INLINE unsigned bits_ctz64(uint64_t v) { return PT_SCAST(unsigned,v)?bits_ctz(PT_SCAST(unsigned,v)):(32U+bits_ctz(PT_SCAST(unsigned,v>>32))); }
  PT_INLINE unsigned bits_ffs64(uint64_t v) { return v?bits_ctz64(v)+1U:0U; }
 #endif

 #define BITS_SWAP16_DEFINED_
 #define BITS_SWAP_DEFINED_
 #ifndef RtlUshortByteSwap
  unsigned short __cdecl _byteswap_ushort(unsigned short);
  #pragma intrinsic(_byteswap_ushort)
  unsigned long __cdecl _byteswap_ulong(unsigned long);
  #pragma intrinsic(_byteswap_ulong)
  uint64_t __cdecl _byteswap_uint64(uint64_t);
  #pragma intrinsic(_byteswap_uint64)
 #endif
 PT_INLINE uint16_t bits_swap16(uint16_t v) { return _byteswap_ushort(v); }
 PT_INLINE unsigned bits_swap32(unsigned v) { return _byteswap_ulong(v); }
 PT_INLINE uint64_t bits_swap64(uint64_t v) { return _byteswap_uint64(v); }

 #if ! defined _M_ARM64 && ! defined _M_ARM64EC
  #define BITS_UMUL_DEFINED_
  #ifndef __INTRIN_H_
   unsigned __int64 __cdecl __emulu(unsigned,unsigned);
   #pragma intrinsic(__emulu)
  #endif
  PT_INLINE unsigned bits_umul(unsigned a,unsigned b,void* h) { uint64_t p = __emulu(a,b); *PT_SCASTP(unsigned*,h) = PT_SCAST(unsigned,p>>32); return PT_SCAST(unsigned,p); }
  PT_INLINE unsigned bits_usqr(unsigned a,void* h) { uint64_t p = __emulu(a,a); *PT_SCASTP(unsigned*,h) = PT_SCAST(unsigned,p>>32); return PT_SCAST(unsigned,p); }
  #if _MSC_VER >= 1928
   #define BITS_UDIV_DEFINED_
   #ifndef __INTRIN_H_
    unsigned _udiv64(unsigned __int64,unsigned,unsigned*);
    #pragma intrinsic(_udiv64)
   #endif
   PT_INLINE unsigned bits_udiv(unsigned nl,unsigned nh,unsigned d) { unsigned r; return _udiv64((PT_SCAST(uint64_t,nh)<<32)|nl,d,&r); }
   PT_INLINE unsigned bits_umod(unsigned nl,unsigned nh,unsigned d) { unsigned r; _udiv64((PT_SCAST(uint64_t,nh)<<32)|nl,d,&r); return r; }
   PT_INLINE unsigned bits_udivmod(unsigned nl,unsigned nh,unsigned d,void* r) { return _udiv64((PT_SCAST(uint64_t,nh)<<32)|nl,d,PT_SCASTP(unsigned*,r)); }
  #endif
  #if UINTPTR_MAX > UINT_MAX
   #define BITS_UMUL64_DEFINED_
   #ifndef __INTRIN_H_
    unsigned __int64 _umul128(unsigned __int64,unsigned __int64,unsigned __int64*);
    #pragma intrinsic(_umul128)
   #endif
   PT_INLINE uint64_t bits_umul64(uint64_t a,uint64_t b,void* h) { return _umul128(a,b,PT_SCASTP(unsigned __int64*,h)); }
   PT_INLINE uint64_t bits_usqr64(uint64_t a,void* h) { return _umul128(a,a,PT_SCASTP(unsigned __int64*,h)); }
   #if _MSC_VER >= 1928
    #define BITS_UDIV64_DEFINED_
    #ifndef __INTRIN_H_
     unsigned __int64 _udiv128(unsigned __int64,unsigned __int64,unsigned __int64,unsigned __int64*);
     #pragma intrinsic(_udiv128)
    #endif
    PT_INLINE uint64_t bits_udiv64(uint64_t nl,uint64_t nh,uint64_t d) { uint64_t r; return _udiv128(nh,nl,d,&r); }
    PT_INLINE uint64_t bits_umod64(uint64_t nl,uint64_t nh,uint64_t d) { uint64_t r; _udiv128(nh,nl,d,&r); return r; }
    PT_INLINE uint64_t bits_udivmod64(uint64_t nl,uint64_t nh,uint64_t d,void* r) { return _udiv128(nh,nl,d,PT_SCASTP(unsigned __int64*,r)); }
  #endif
  #endif
 #endif
 #if defined _M_ARM64 || defined _M_ARM64EC
  #define BITS_UMUL64_DEFINED_
  #ifndef __INTRIN_H_
   unsigned __int64 __umulh(unsigned __int64,unsigned __int64);
   #pragma intrinsic(__umulh)
  #endif
  PT_INLINE uint64_t bits_umul64(uint64_t a,uint64_t b,void* h) { *PT_SCASTP(uint64_t*,h) = __umulh(a,b); return a*b; }
  PT_INLINE uint64_t bits_usqr64(uint64_t a,void* h) { *PT_SCASTP(uint64_t*,h) = __umulh(a,a); return a*a; }
 #endif

#else

 #define BITS_CLZ_DEFINED_
 PT_INLINE unsigned bits_clz( unsigned v) { return v ? PT_SCAST( unsigned, __builtin_clz( v)) : PT_SCAST( unsigned, sizeof( v) * 8); }
 PT_INLINE unsigned bits_fls( unsigned v) { return v ? PT_SCAST( unsigned, sizeof( v) * 8) - PT_SCAST( unsigned, __builtin_clz( v)) : 0u; }
 PT_INLINE unsigned bits_ctz( unsigned v) { return v ? PT_SCAST( unsigned, __builtin_ctz( v)) : PT_SCAST( unsigned, sizeof( v) * 8); }
 PT_INLINE unsigned bits_ffs( unsigned v) { return v ? PT_SCAST( unsigned, __builtin_ctz( v)) + 1 : 0u; }
 #if UINTPTR_MAX > UINT_MAX
  PT_INLINE unsigned bits_clz64( uint64_t v) { return v ? PT_SCAST( unsigned, __builtin_clzl( v)) : sizeof( v) * 8; }
  PT_INLINE unsigned bits_fls64( uint64_t v) { return v ? PT_SCAST( unsigned, sizeof( v) * 8) - PT_SCAST( unsigned, __builtin_clzl( v)) : 0u; }
  PT_INLINE unsigned bits_ctz64( uint64_t v) { return v ? PT_SCAST( unsigned, __builtin_ctzl( v)) : sizeof( v) * 8; }
  PT_INLINE unsigned bits_ffs64( uint64_t v) { return v ? PT_SCAST( unsigned, __builtin_ctzl( v)) + 1 : 0u; }
 #else
  PT_INLINE unsigned bits_clz64( uint64_t v) { unsigned h = PT_SCAST( unsigned, v >> 32); return h ? PT_SCAST( unsigned, __builtin_clz( h)) : (PT_SCAST( unsigned, v) ? PT_SCAST( unsigned, __builtin_clz( PT_SCAST( unsigned, v))) + 32 : 64u); }
  PT_INLINE unsigned bits_fls64( uint64_t v) { return v ? PT_SCAST( unsigned, sizeof( v) * 8) - bits_clz64( v) : 0u; }
  PT_INLINE unsigned bits_ctz64( uint64_t v) { unsigned h; return PT_SCAST( unsigned, v) ? PT_SCAST( unsigned, __builtin_ctz( PT_SCAST( unsigned, v))) : (h = PT_SCAST( unsigned, v >> 32), h ? PT_SCAST( unsigned, __builtin_ctz( h)) + 32 : 64u); }
  PT_INLINE unsigned bits_ffs64( uint64_t v) { return v ? bits_ctz64( v) + 1 : 0u; }
 #endif
 #define BITS_SWAP_DEFINED_
 PT_INLINE unsigned bits_swap32(unsigned v) { return PT_SCAST(unsigned,__builtin_bswap32(PT_SCAST(int,v))); }
 PT_INLINE uint64_t bits_swap64(uint64_t v) { return PT_SCAST(uint64_t,__builtin_bswap64(PT_SCAST(long long,v))); }
 #if defined(__GNUC__) && !defined(INTEL_CC)
  #if defined(__i386__) || defined(__x86_64__)
   #define BITS_ROT_DEFINED_
   PT_INLINE unsigned bits_ror(unsigned v,unsigned c) { __asm__("rorl %%cl,%0": "=r" (v): "0" (v), "c" (c)); return v; }
   PT_INLINE unsigned bits_rol(unsigned v,unsigned c) { __asm__("roll %%cl,%0": "=r" (v): "0" (v), "c" (c)); return v; }
   #if UINTPTR_MAX > UINT_MAX
    #define BITS_ROT64_DEFINED_
    PT_INLINE uintptr_t bits_ror64(uint64_t v,unsigned c) { __asm__("rorq %%cl,%0": "=r" (v): "0" (v), "c" (c)); return v; }
    PT_INLINE uintptr_t bits_rol64(uint64_t v,unsigned c) { __asm__("rolq %%cl,%0": "=r" (v): "0" (v), "c" (c)); return v; }
   #endif
  #endif
 #endif

#endif

  // Majority of (if not all) modern general purpose compilers
  // implement arithmetic shift on signed values, regardless
  // of what latitude the C spec gives them.
PT_INLINE int bits_sar(int v,unsigned c)            { return v>>c; }
PT_INLINE int64_t bits_sar64(int64_t v,unsigned c)  { return v>>c; }
PT_INLINE int bits_sal(int v,unsigned c)            { return v<<c; }
PT_INLINE int64_t bits_sal64(int64_t v,unsigned c)  { return v<<c; }
#ifndef BITS_ROT_DEFINED_
 PT_INLINE unsigned bits_ror(unsigned v,unsigned c) { return bits_ror_slow(v,c); }
 PT_INLINE unsigned bits_rol(unsigned v,unsigned c) { return bits_rol_slow(v,c); }
#endif
#ifndef BITS_ROT64_DEFINED_
 PT_INLINE uint64_t bits_ror64(uint64_t v,unsigned c) { return bits_ror64_slow(v,c); }
 PT_INLINE uint64_t bits_rol64(uint64_t v,unsigned c) { return bits_rol64_slow(v,c); }
#endif
#ifndef BITS_CLZ_DEFINED_
 PT_INLINE unsigned bits_clz(unsigned v) { return bits_clz_slow(v); }
 PT_INLINE unsigned bits_fls(unsigned v) { return bits_fls_slow(v); }
 PT_INLINE unsigned bits_ctz(unsigned v) { return bits_ctz_slow(v); }
 PT_INLINE unsigned bits_ffs(unsigned v) { return bits_ffs_slow(v); }
 PT_INLINE unsigned bits_clz64(uint64_t v) { return bits_clz64_slow(v); }
 PT_INLINE unsigned bits_fls64(uint64_t v) { return bits_fls64_slow(v); }
 PT_INLINE unsigned bits_ctz64(uint64_t v) { return bits_ctz64_slow(v); }
 PT_INLINE unsigned bits_ffs64(uint64_t v) { return bits_ffs64_slow(v); }
#endif
#ifndef BITS_POP_DEFINED_
 PT_INLINE unsigned bits_pop(unsigned v) { return bits_pop_slow(v); }
 PT_INLINE unsigned bits_pop64(uint64_t v) { return bits_pop64_slow(v); }
#endif
#ifndef BITS_SWAP16_DEFINED_
 PT_INLINE uint16_t bits_swap16(uint16_t v) { return bits_swap16_slow(v); }
#endif
#ifndef BITS_SWAP_DEFINED_
 PT_INLINE unsigned bits_swap32(unsigned v) { return bits_swap32_slow(v); }
 PT_INLINE uint64_t bits_swap64(uint64_t v) { return bits_swap64_slow(v); }
#endif
#ifndef BITS_UMUL_DEFINED_
 PT_INLINE unsigned bits_umul(unsigned a,unsigned b,void* h) { return bits_umul_slow(a,b,h); }
 PT_INLINE unsigned bits_usqr(unsigned a,void* h) { return bits_usqr_slow(a,h); }
#endif
#ifndef BITS_UMUL64_DEFINED_
 PT_INLINE uint64_t bits_umul64(uint64_t a,uint64_t b,void* h) { return bits_umul64_slow(a,b,h); }
 PT_INLINE uint64_t bits_usqr64(uint64_t a,void* h) { return bits_usqr64_slow(a,h); }
#endif
#ifndef BITS_UDIV_DEFINED_
  PT_INLINE unsigned bits_udiv(unsigned nl,unsigned nh,unsigned d) { return bits_udiv_slow(nl,nh,d); }
  PT_INLINE unsigned bits_umod(unsigned nl,unsigned nh,unsigned d) { return bits_umod_slow(nl,nh,d); }
  PT_INLINE unsigned bits_udivmod(unsigned nl,unsigned nh,unsigned d,void* r) { return bits_udivmod_slow(nl,nh,d,r); }
#endif
#ifndef BITS_UDIV64_DEFINED_
  PT_INLINE uint64_t bits_udiv64(uint64_t nl,uint64_t nh,uint64_t d) { return bits_udiv64_slow(nl,nh,d); }
  PT_INLINE uint64_t bits_umod64(uint64_t nl,uint64_t nh,uint64_t d) { return bits_umod64_slow(nl,nh,d); }
  PT_INLINE uint64_t bits_udivmod64(uint64_t nl,uint64_t nh,uint64_t d,void* r) { return bits_udivmod64_slow(nl,nh,d,r); }
#endif

#if UINTPTR_MAX > UINT_MAX
 PT_INLINE uintptr_t  bits_rorp_slow(uintptr_t v,unsigned c) { return bits_ror64_slow(v,c); }
 PT_INLINE uintptr_t  bits_rolp_slow(uintptr_t v,unsigned c) { return bits_rol64_slow(v,c); }
 PT_INLINE unsigned  bits_clzp_slow(uintptr_t v)           { return bits_clz64_slow(v); }
 PT_INLINE unsigned  bits_ctzp_slow(uintptr_t v)           { return bits_ctz64_slow(v); }
 PT_INLINE unsigned  bits_flsp_slow(uintptr_t v)           { return bits_fls64_slow(v); }
 PT_INLINE unsigned  bits_ffsp_slow(uintptr_t v)           { return bits_ffs64_slow(v); }
 PT_INLINE unsigned  bits_popp_slow(uintptr_t v)           { return bits_pop64_slow(v); }
 PT_INLINE uintptr_t bits_swapp_slow(uintptr_t v)          { return bits_swap64_slow(v); }
 PT_INLINE uintptr_t bits_umulp_slow(uintptr_t a,uintptr_t b,void* h) { return bits_umul64_slow(a,b,h); }
 PT_INLINE uintptr_t bits_usqrp_slow(uintptr_t a,void* h) { return bits_usqr64_slow(a,h); }
 PT_INLINE intptr_t  bits_sarp(intptr_t v,unsigned c) { return bits_sar64(v,c); }
 PT_INLINE intptr_t  bits_salp(intptr_t v,unsigned c) { return bits_sal64(v,c); }
 PT_INLINE uintptr_t  bits_rorp(uintptr_t v,unsigned c) { return bits_ror64(v,c); }
 PT_INLINE uintptr_t  bits_rolp(uintptr_t v,unsigned c) { return bits_rol64(v,c); }
 PT_INLINE unsigned  bits_clzp(uintptr_t v)           { return bits_clz64(v); }
 PT_INLINE unsigned  bits_ctzp(uintptr_t v)           { return bits_ctz64(v); }
 PT_INLINE unsigned  bits_flsp(uintptr_t v)           { return bits_fls64(v); }
 PT_INLINE unsigned  bits_ffsp(uintptr_t v)           { return bits_ffs64(v); }
 PT_INLINE unsigned  bits_popp(uintptr_t v)           { return bits_pop64(v); }
 PT_INLINE uintptr_t bits_swapp(uintptr_t v)          { return bits_swap64(v); }
 PT_INLINE uintptr_t bits_umulp(uintptr_t a,uintptr_t b,void* h) { return bits_umul64(a,b,h); }
 PT_INLINE uintptr_t bits_usqrp(uintptr_t a,void* h) { return bits_usqr64(a,h); }
 PT_INLINE uintptr_t bits_udivp(uintptr_t nl,uintptr_t nh,uintptr_t d) { return bits_udiv64(nl,nh,d); }
 PT_INLINE uintptr_t bits_umodp(uintptr_t nl,uintptr_t nh,uintptr_t d) { return bits_umod64(nl,nh,d); }
 PT_INLINE uintptr_t bits_udivmodp(uintptr_t nl,uintptr_t nh,uintptr_t d,void* r) { return bits_udivmod64(nl,nh,d,r); }
#else
 PT_INLINE uintptr_t  bits_rorp_slow(uintptr_t v,unsigned c) { return bits_ror_slow(v,c); }
 PT_INLINE uintptr_t  bits_rolp_slow(uintptr_t v,unsigned c) { return bits_rol_slow(v,c); }
 PT_INLINE unsigned  bits_clzp_slow(uintptr_t v)           { return bits_clz_slow(v); }
 PT_INLINE unsigned  bits_ctzp_slow(uintptr_t v)           { return bits_ctz_slow(v); }
 PT_INLINE unsigned  bits_flsp_slow(uintptr_t v)           { return bits_fls_slow(v); }
 PT_INLINE unsigned  bits_ffsp_slow(uintptr_t v)           { return bits_ffs_slow(v); }
 PT_INLINE unsigned  bits_popp_slow(uintptr_t v)           { return bits_pop_slow(v); }
 PT_INLINE uintptr_t bits_swapp_slow(uintptr_t v)          { return bits_swap32_slow(v); }
 PT_INLINE uintptr_t bits_umulp_slow(uintptr_t a,uintptr_t b,void* h) { return bits_umul_slow(a,b,h); }
 PT_INLINE uintptr_t bits_usqrp_slow(uintptr_t a,void* h) { return bits_usqr_slow(a,h); }
 PT_INLINE intptr_t  bits_sarp(intptr_t v,unsigned c) { return bits_sar(v,c); }
 PT_INLINE intptr_t  bits_salp(intptr_t v,unsigned c) { return bits_sal(v,c); }
 PT_INLINE uintptr_t  bits_rorp(uintptr_t v,unsigned c) { return bits_ror(v,c); }
 PT_INLINE uintptr_t  bits_rolp(uintptr_t v,unsigned c) { return bits_rol(v,c); }
 PT_INLINE unsigned  bits_clzp(uintptr_t v)           { return bits_clz(v); }
 PT_INLINE unsigned  bits_ctzp(uintptr_t v)           { return bits_ctz(v); }
 PT_INLINE unsigned  bits_flsp(uintptr_t v)           { return bits_fls(v); }
 PT_INLINE unsigned  bits_ffsp(uintptr_t v)           { return bits_ffs(v); }
 PT_INLINE unsigned  bits_popp(uintptr_t v)           { return bits_pop(v); }
 PT_INLINE uintptr_t bits_swapp(uintptr_t v)          { return bits_swap32(v); }
 PT_INLINE uintptr_t bits_umulp(uintptr_t a,uintptr_t b,void* h) { return bits_umul(a,b,h); }
 PT_INLINE uintptr_t bits_usqrp(uintptr_t a,void* h) { return bits_usqr(a,h); }
 PT_INLINE uintptr_t bits_udivp(uintptr_t nl,uintptr_t nh,uintptr_t d) { return bits_udiv(nl,nh,d); }
 PT_INLINE uintptr_t bits_umodp(uintptr_t nl,uintptr_t nh,uintptr_t d) { return bits_umod(nl,nh,d); }
 PT_INLINE uintptr_t bits_udivmodp(uintptr_t nl,uintptr_t nh,uintptr_t d,void* r) { return bits_udivmod(nl,nh,d,r); }
#endif

   // Hash functions based on Bob Jenkins lookup3.c
#define BITS_HASH_MIX(a,b,c) do { a -= c; a ^= bits_rol(c, 4); c += b; b -= a; b ^= bits_rol(a, 6); a += c; c -= b; c ^= bits_rol(b, 8); b += a; a -= c; a ^= bits_rol(c,16); c += b; b -= a; b ^= bits_rol(a,19); a += c; c -= b; c ^= bits_rol(b, 4); b += a; } while(0)
#define BITS_HASH_FINAL(a,b,c) do { c ^= b; c -= bits_rol(b,14); a ^= c; a -= bits_rol(c,11); b ^= a; b -= bits_rol(a,25); c ^= b; c -= bits_rol(b,16); a ^= c; a -= bits_rol(c,4); b ^= a; b -= bits_rol(a,14); c ^= b; c -= bits_rol(b,24); } while(0)
   // Hash functions based on Thomas Wang "Integer Hash Function" paper ca 1997.
#define BITS_HASH_UINT(a) do { a = ~a+(a<<15); a = a^(a>>12); a = a+(a<<2); a = a^(a>>4); a = a+(a<<3)+(a<<11); a = a^(a>>16); } while(0)
#if UINTPTR_MAX > UINT_MAX
#define BITS_HASH_UINT64(a) do { a = ~a+(a<<18); a = a^(a>>31); a = a+(a<<2)+(a<<4); a = a^(a>>11); a = a+(a<<6); a = a^(a>>22); } while(0)
#endif

PT_INLINE unsigned bits_hash_uint(unsigned k)
{
   unsigned a = k;
   BITS_HASH_UINT(a);
   return a;
}

PT_INLINE unsigned bits_hash_uint64(uint64_t k)
{
#if UINTPTR_MAX > UINT_MAX
   uint64_t a = k;
   BITS_HASH_UINT64(a);
   return PT_SCAST(unsigned,a);
#else
   unsigned a = PT_SCAST(unsigned,k);
   unsigned b = PT_SCAST(unsigned,k>>32);
   unsigned c = 0;
   BITS_HASH_FINAL(a,b,c);
   return c;
#endif
}

PT_INLINE unsigned bits_hash_int( int k) { return bits_hash_uint( PT_SCAST( unsigned, k)); }
PT_INLINE unsigned bits_hash_int64( int64_t k) { return bits_hash_uint64( PT_SCAST( uint64_t, k)); }
#if UINTPTR_MAX > UINT_MAX
PT_INLINE unsigned bits_hash_size( uintptr_t k) { return bits_hash_uint64( k); }
#else
PT_INLINE unsigned bits_hash_size( uintptr_t k) { return bits_hash_uint( k); }
#endif
PT_INLINE unsigned bits_hash_ptr(const void* k) { return bits_hash_size( PT_RCAST( uintptr_t, k)); }

uint64_t PT_CCALL bits_des_mix64(uint64_t input);

#undef BITS_ROT_DEFINED_
#undef BITS_ROT64_DEFINED_
#undef BITS_CLZ_DEFINED_
#undef BITS_POP_DEFINED_
#undef BITS_SWAP16_DEFINED_
#undef BITS_SWAP_DEFINED_
#undef BITS_UMUL_DEFINED_
#undef BITS_UMUL64_DEFINED_
#undef BITS_UDIV_DEFINED_
#undef BITS_UDIV64_DEFINED_

PT_EXTERNC_END
#endif
