//---------------------------------------------------------------------------
// Copyright 1997-2022 Joe Lowe
//
// Permission is granted to any person obtaining a copy of this Software,
// to deal in the Software without restriction, including the rights to use,
// copy, modify, merge, publish, distribute, sublicense, and sell copies of
// the Software.
//
// The above copyright and permission notice must be left intact in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS WITHOUT WARRANTY.
//---------------------------------------------------------------------------
// file name:  ptlibc.h
// created:    2013.01.30
//    Include base C run time headers, or replacements for
//    lagging environments.
//---------------------------------------------------------------------------
#ifndef PTLIBC_H
#define PTLIBC_H
#include "ptcompiler.h"
PT_EXTERNC_START

#if defined(_MSC_VER)
   #ifndef UNICODE
   #define UNICODE
   #endif
   #ifndef _UNICODE
   #define _UNICODE
   #endif
#endif

#if !defined(KERNEL)

#ifdef __cplusplus
} // extern "C"
   // The C++ headers undefs a bunch of macros,
   // including "free" used for heap debugging. Include
   // the header now so it will not break things if
   // included again later.
#include <cstdlib>
#ifdef _LIBCPP_FALLTHROUGH
#undef _LIBCPP_FALLTHROUGH
#define _LIBCPP_FALLTHROUGH() PT_FALLTHROUGH
#endif
#endif
#include <stdarg.h>
#include <stddef.h>
#include <limits.h>
#include <sys/types.h>
#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <wchar.h>
#include <memory.h>
#ifndef __cplusplus
#include <stdbool.h>
#else
#include <new>
extern "C" {
#endif
   // Some inline Windows system header code will generate
   // dependencies on new msvcrt.dll.
#undef __GOT_SECURE_LIB__
#undef __STDC_SECURE_LIB__

#else // KERNEL

#ifdef _WIN32

#define malloc hide_malloc
#define calloc hide_calloc
#define free hide_free
#ifdef __cplusplus
} // extern "C"
#endif
#include <stdarg.h>
#include <stddef.h>
#include <limits.h>
#include <memory.h>
#include <wchar.h>
#include "ptstdint.h"
#ifndef __cplusplus
#include <stdbool.h>
#else
extern "C" {
#endif
#undef malloc
#undef calloc
#undef free
void PT_CCALL module_initialize(void);
void PT_CCALL module_finalize(int/*bool*/ unloading);

#elif defined(__APPLE__)

#ifdef __cplusplus
} // extern "C"
#endif
#ifndef __cplusplus
typedef __WCHAR_TYPE__ wchar_t;
#endif
#include <sys/types.h>
#include <stdarg.h>
#include <machine/limits.h>
#include <stdint.h>
#include <stdbool.h>
#if PTRDIFF_MAX > INT_MAX && SIZE_MAX == UINT_MAX
   // OSX 10.6 SDK has broken SIZE_MAX in stdint.h
#undef SIZE_MAX
#define SIZE_MAX UINT64_MAX
#endif
#if SIZE_MAX > UINT_MAX
typedef long ptrdiff_t;
#else
typedef int ptrdiff_t;
#endif
#include "ptwchar.h"
#ifdef __cplusplus
extern "C" {
#endif
#include <libkern/libkern.h>
   // OSX kernel offsetof macro is not good enough for GCC.
#undef offsetof
#define offsetof(type, field) __builtin_offsetof(type, field)
int/*err*/ PT_CCALL module_initialize(void*,void*);
void PT_CCALL module_finalize(void*,void*);
#undef memcpy
#undef memmove
#undef strcpy
#undef strncpy
#undef strcat
#undef strcmp

#elif defined(__linux__)

void PT_CCALL kpi_simd_begin(void);
void PT_CCALL kpi_simd_end(void);
#define KPI_REALIGN_STACK __attribute__((force_align_arg_pointer))

#ifdef LINUX_VER

   // Linux defines 16 bit wchar_t, but with possibly mismatched typeof(L'').
#define wchar_t wchar16_t
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/types.h>
#undef wchar_t
#ifndef _TIME_T
typedef long time_t;
#endif
PT_C_ASSERT1(sizeof(time_t) == sizeof(long));
PT_C_ASSERT1(sizeof(uid_t) == sizeof(int));
PT_C_ASSERT1(sizeof(gid_t) == sizeof(int));

#define KPI_SIMD_BEGIN kernel_fpu_begin(); {
#define KPI_SIMD_END } kernel_fpu_end();

#else

   // Can not use kernel headers for portable code being
   // compiled for use in the linux kernel.
#include <stdarg.h>
typedef long time_t;
typedef int uid_t;
typedef int gid_t;
#if defined(__x86_64__) || defined(__powerpc64__) || defined(__ppc64__)
#define BITS_PER_LONG 64
typedef unsigned long size_t;
typedef long ssize_t;
typedef long ptrdiff_t;
#else
#define BITS_PER_LONG 32
typedef unsigned size_t;
typedef int ssize_t;
typedef int ptrdiff_t;
#endif
#define offsetof(type, field) __builtin_offsetof(type, field)
#define NULL ((void*)0)

void* PT_CCALL ptbase_memcpy(void*PT_RESTRICT a,const void*PT_RESTRICT b,size_t c);
void* PT_CCALL ptbase_memset(void*PT_RESTRICT a,int b,size_t c);
void* PT_CCALL ptbase_memmove(void* a,const void* b,size_t c);
int PT_CCALL ptbase_memcmp(const void* a,const void* b,size_t c);
size_t PT_CCALL ptbase_strlen(const char* a);
char* PT_CCALL ptbase_strcpy(char*PT_RESTRICT a,const char*PT_RESTRICT b);
char* PT_CCALL ptbase_strcat(char*PT_RESTRICT a,const char*PT_RESTRICT b);
int PT_CCALL ptbase_strcmp(const char* a,const char* b);
PT_INLINE void* memcpy(void*PT_RESTRICT a,const void*PT_RESTRICT b,size_t c) { return ptbase_memcpy(a,b,c); }
PT_INLINE void* memset(void*PT_RESTRICT a,int b,size_t c) { return ptbase_memset(a,b,c); }
PT_INLINE void* memmove(void* a,const void* b,size_t c) { return ptbase_memmove(a,b,c); }
PT_INLINE int memcmp(const void* a,const void* b,size_t c) { return ptbase_memcmp(a,b,c); }
PT_INLINE size_t strlen(const char* a) { return ptbase_strlen(a); }
PT_INLINE char* strcpy(char*PT_RESTRICT a,const char*PT_RESTRICT b) { return ptbase_strcpy(a,b); }
PT_INLINE char* strcat(char*PT_RESTRICT a,const char*PT_RESTRICT b) { return ptbase_strcat(a,b); }
PT_INLINE int strcmp(const char* a,const char* b) { return ptbase_strcmp(a,b); }

#define KPI_SIMD_BEGIN kpi_simd_begin(); {
#define KPI_SIMD_END } kpi_simd_end();

#endif

#if __WCHAR_MAX__ < 0x7FFFFFFF
typedef int wchar_t;
#else
typedef __WCHAR_TYPE__ wchar_t;
#endif
#include "ptlimits.h"
#include "ptstdint.h"
#undef bool
#define bool _Bool
#undef true
#define true 1
#undef false
#define false 0

void PT_CCALL module_initialize(void);
void PT_CCALL module_finalize(void);

#endif

#endif // KERNEL

#ifndef va_copy
   #ifdef __va_copy
      #define va_copy(d,s) __va_copy(d,s)
      #define va_copyend(d) va_end(d)
   #else
      #define va_copy(d,s) ((d) = (s))
      #define va_copyend(d)
   #endif
#elif !defined(va_copyend)
   #define va_copyend(d) va_end(d)
#endif

#ifndef INTPTR_C
#if SIZE_MAX == UINT_MAX
#define INTPTR_C(c)  c
#define UINTPTR_C(c) c##U
#else
#define INTPTR_C(c)  INT64_C(c)
#define UINTPTR_C(c) UINT64_C(c)
#endif
#endif
#ifndef SIZE_C
#if SIZE_MAX == UINT_MAX
#define SIZE_C(c) c##U
#else
#define SIZE_C(c) UINT64_C(c)
#endif
#endif
#ifndef PTRDIFF_C
#if SIZE_MAX == UINT_MAX
#define PTRDIFF_C(c) c
#else
#define PTRDIFF_C(c) INT64_C(c)
#endif
#endif

#ifndef __cplusplus
#undef nullptr
#define nullptr ((void*)0)
#endif

PT_C_ASSERT1(sizeof(void*) == sizeof(size_t));
PT_C_ASSERT1((sizeof(size_t) == 4 && SIZE_MAX == 0xFFFFFFFF) || (sizeof(size_t) == 8 && SIZE_MAX == UINT64_C(0xFFFFFFFFFFFFFFFF)));
PT_C_ASSERT1(sizeof(void*) == sizeof(ptrdiff_t));
PT_C_ASSERT1((sizeof(ptrdiff_t) == 4 && PTRDIFF_MAX == 0x7FFFFFFF && PTRDIFF_MIN+1+0x7FFFFFFF == 0) || (sizeof(ptrdiff_t) == 8 && PTRDIFF_MAX == INT64_C(0x7FFFFFFFFFFFFFFF) && PTRDIFF_MIN+1+PTRDIFF_MAX == 0));
PT_C_ASSERT1(sizeof(void*) == sizeof(uintptr_t));
PT_C_ASSERT1((sizeof(uintptr_t) == 4 && UINTPTR_MAX == 0xFFFFFFFF) || (sizeof(uintptr_t) == 8 && UINTPTR_MAX == UINT64_C(0xFFFFFFFFFFFFFFFF)));
PT_C_ASSERT1(sizeof(void*) == sizeof(intptr_t));
PT_C_ASSERT1((sizeof(intptr_t) == 4 && INTPTR_MAX == 0x7FFFFFFF && INTPTR_MIN+1+0x7FFFFFFF == 0) || (sizeof(intptr_t) == 8 && INTPTR_MAX == INT64_C(0x7FFFFFFFFFFFFFFF) && INTPTR_MIN+1+INTPTR_MAX == 0));
PT_C_ASSERT1(sizeof(wchar_t) != 2 || (WCHAR_MIN == 0 && WCHAR_MAX == 0xFFFF));
PT_C_ASSERT1(sizeof(wchar_t) != 4 || (WCHAR_MIN == INT_MIN && WCHAR_MAX == INT_MAX) || (WCHAR_MAX == UINT_MAX));
#ifndef LINUX_VER
PT_C_ASSERT1(sizeof(L"") == sizeof(wchar_t));
#endif
#ifndef _WIN32
PT_C_ASSERT1(sizeof(time_t) == sizeof(long));
#endif

#undef PT_BOOL8
#undef PT_INT8
#undef PT_UINT8
#undef PT_INT16
#undef PT_UINT16
#undef PT_INT64
#undef PT_UINT64
#undef PT_SIZE_T
#if PT_SIZEOF_BOOL == 1
#define PT_BOOL8 bool
#else
#define PT_BOOL8 uint8_t
#endif
#define PT_INT8 int8_t
#define PT_UINT8 uint8_t
#define PT_INT16 int16_t
#define PT_UINT16 uint16_t
#define PT_INT64 int64_t
#define PT_UINT64 uint64_t
#define PT_SIZE_T size_t
typedef union PT_ALIGNAS(16) pt_vec128_t { PT_UINT8 u8[16]; PT_UINT16 u16[8]; PT_UINT32 u32[4]; PT_UINT64 u64[2]; } pt_vec128_t;
#define PT_VEC128 pt_vec128_t

#define pt_t_malloc(t)    PT_SCAST(t*,malloc(sizeof(t)))
#define pt_tn_malloc(t,n) PT_SCAST(t*,malloc((n)*sizeof(t)))
#define pt_ts_malloc(t,s) PT_SCAST(t*,malloc(s))
#define pt_t_calloc(t)    PT_SCAST(t*,calloc(1,sizeof(t)))
#define pt_tn_calloc(t,n) PT_SCAST(t*,calloc(n,sizeof(t)))
#define pt_ts_calloc(t,s) PT_SCAST(t*,calloc(1,s))

#ifdef __cplusplus
#define pt_t_memzero_(t,d) do{memset(d,0,t);}while(0)
#define pt_tn_memzero_(t,d,n) do{memset(d,0,(n)*(t));}while(0)
#define pt_ts_memzero_(d,n) do{memset(d,0,n);}while(0)
#define pt_t_memcpy_(t,d,s) do{memcpy(d,s,t);}while(0)
#define pt_tn_memcpy_(t,d,s,n) do{memcpy(d,s,(n)*(t));}while(0)
#define pt_tn_memmove_(t,d,s,n) do{memmove(d,s,(n)*(t));}while(0)
#else
#define pt_t_memzero_(t,d) do{memset((void*)(d),0,t);}while(0)
#define pt_tn_memzero_(t,d,n) do{memset((void*)(d),0,(n)*(t));}while(0)
#define pt_ts_memzero_(d,n) do{memset((void*)(d),0,n);}while(0)
#define pt_t_memcpy_(t,d,s) do{memcpy((void*)(d),s,t);}while(0)
#define pt_tn_memcpy_(t,d,s,n) do{memcpy((void*)(d),s,(n)*(t));}while(0)
#define pt_tn_memmove_(t,d,s,n) do{memmove((void*)(d),s,(n)*(t));}while(0)
#endif
#if defined(__cplusplus) || !defined(DEBUG)
#define pt_t_memzero(d) pt_t_memzero_(sizeof(*(d)),d)
#define pt_tn_memzero(d,n) pt_tn_memzero_(sizeof(*(d)),d,n)
#define pt_tN_memzero(d) pt_tn_memzero_(sizeof((d)[0]),d,PT_C_ARRAYLEN(d))
#define pt_t_memcpy(d,s) pt_t_memcpy_(sizeof(*(d)),d,s)
#define pt_tn_memcpy(d,s,n) pt_tn_memcpy_(sizeof(*(d)),d,s,n)
#define pt_tn2_memcpy(d,s,n) pt_tn_memcpy_(sizeof(*(s)),d,s,n)
#define pt_tN_memcpy(d,s) pt_tn_memcpy_(sizeof((d)[0]),d,s,PT_C_ARRAYLEN(d))
#define pt_tN2_memcpy(d,s) pt_tn_memcpy_(sizeof((s)[0]),d,s,PT_C_ARRAYLEN(s))
#define pt_tn_memmove(d,s,n) pt_tn_memmove_(sizeof(*(d)),d,s,n)
#define pt_tn2_memmove(d,s,n) pt_tn_memmove_(sizeof(*(s)),d,s,n)
#define pt_tN_memmove(d,s) pt_tn_memmove_(sizeof((d)[0]),d,s,PT_C_ARRAYLEN(d))
#define pt_tN2_memmove(d,s) pt_tn_memmove_(sizeof((s)[0]),d,s,PT_C_ARRAYLEN(s))
#define pt_ts_memzero(d,n) pt_ts_memzero_(d,n)
#else
#define pt_t_memzero(d) do{if(0) *(d)=*(d);else pt_t_memzero_(sizeof(*(d)),d);}while(0)
#define pt_tn_memzero(d,n) do{if(0) *(d)=*(d);else pt_tn_memzero_(sizeof(*(d)),d,n);}while(0)
#define pt_tN_memzero(d) do{if(0) *(d)=*(d);else pt_tn_memzero_(sizeof((d)[0]),d,PT_C_ARRAYLEN(d));}while(0)
#define pt_t_memcpy(d,s) do{if(0) *(d)=*(d);else pt_t_memcpy_(sizeof(*(d)),d,s);}while(0)
#define pt_tn_memcpy(d,s,n) do{if(0) *(d)=*(d);else pt_tn_memcpy_(sizeof(*(d)),d,s,n);}while(0)
#define pt_tn2_memcpy(d,s,n) do{if(0) *(d)=*(d);else pt_tn_memcpy_(sizeof(*(s)),d,s,n);}while(0)
#define pt_tN_memcpy(d,s) do{if(0) *(d)=*(d);else pt_tn_memcpy_(sizeof((d)[0]),d,s,PT_C_ARRAYLEN(d));}while(0)
#define pt_tN2_memcpy(d,s) do{if(0) *(d)=*(d);else pt_tn_memcpy_(sizeof((s)[0]),d,s,PT_C_ARRAYLEN(s));}while(0)
#define pt_tn_memmove(d,s,n) do{if(0) *(d)=*(d);else pt_tn_memmove_(sizeof(*(d)),d,s,n);}while(0)
#define pt_tn2_memmove(d,s,n) do{if(0) *(d)=*(d);else pt_tn_memmove_(sizeof(*(s)),d,s,n);}while(0)
#define pt_tN_memmove(d,s) do{if(0) *(d)=*(d);else pt_tn_memmove_(sizeof((d)[0]),d,s,PT_C_ARRAYLEN(d));}while(0)
#define pt_tN2_memmove(d,s) do{if(0) *(d)=*(d);else pt_tn_memmove_(sizeof((s)[0]),d,s,PT_C_ARRAYLEN(s));}while(0)
#define pt_tn_memzero(d,n) do{if(0) *(d)=*(d);else pt_tn_memzero_(sizeof(*(d)),d,n);}while(0)
#define pt_ts_memzero(d,n) do{if(0) *(uint8_t*)(d)=*(uint8_t*)(d);else pt_ts_memzero_(d,n);}while(0)
#endif

PT_INLINE int pt_t_memcmp_(size_t t,const void* a,const void* b) {return memcmp(a,b,t);}
PT_INLINE int pt_tn_memcmp_(size_t t,const void* a,const void* b,size_t n) {return memcmp(a,b,n*t);}
PT_INLINE int pt_ts_memcmp(const void* d1,size_t s1,const void* d2,size_t s2) {if(s1==s2) return memcmp(d1,d2,s1);if(s1<s2) return(memcmp(d1,d2,s1)>0)*2-1; return(memcmp(d1,d2,s2)>=0)*2-1;}
PT_INLINE_BOOL pt_ts_memiszero(const void* d,size_t s) {PT_UINT8 c=0;PT_SIZE_T i;for(i=0;i<s;i++)c|=PT_SCASTP(const PT_UINT8*,d)[i];return !c;}
#define pt_t_memcmp(a,b) pt_t_memcmp_(sizeof(*(a)),a,b)
#define pt_t2_memcmp(a,b) pt_t_memcmp_(sizeof(*(b)),a,b)
#define pt_tn_memcmp(a,b,n) pt_tn_memcmp_(sizeof(*(a)),a,b,n)
#define pt_tn2_memcmp(a,b,n) pt_tn_memcmp_(sizeof(*(b)),a,b,n)
#define pt_tN_memcmp(a,b) pt_tn_memcmp_(sizeof((a)[0]),a,b,PT_C_ARRAYLEN(a))
#define pt_tN2_memcmp(a,b) pt_tn_memcmp_(sizeof((b)[0]),a,b,PT_C_ARRAYLEN(b))
#define pt_ts_memcmp(a,b,c,d) pt_ts_memcmp(a,b,c,d)
#define pt_t_memiszero(a) pt_ts_memiszero(a,sizeof(*(a)))
#define pt_tn_memiszero(a,n) pt_ts_memiszero(a,sizeof((a)[0])*(n))
#define pt_tN_memiszero(a) pt_ts_memiszero(a,PT_C_ARRAYSIZE(a))

#ifndef KPI_SIMD_BEGIN
#define KPI_REALIGN_STACK
#define KPI_SIMD_BEGIN {
#define KPI_SIMD_END   }
#endif

PT_EXTERNC_END
#endif
